/*
 * Copyright 2000 Hans Reiser, see reiserfs/README for licensing and copyright details
 */

#ifdef __KERNEL__

#include <linux/fs.h>
#include <linux/reiserfs_fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#else

#include "nokernel.h"

#endif


/*
** reiserfs_ioctl - handler for ioctl for inode
** supported commands:
**  1) REISERFS_IOC_UNPACK - try to unpack tail from direct item into indirect
**                           and prevent packing file (argument arg has to be non-zero)
**  2) That's all for a while ...
*/
int reiserfs_ioctl (struct inode * inode, struct file * filp, unsigned int cmd,
		unsigned long arg)
{
	switch (cmd) {
	    case REISERFS_IOC_UNPACK:
		if (arg)
		    return reiserfs_unpack (inode, filp);
			
	    default:
		return -ENOTTY;
	}
}


/*
** reiserfs_unpack
** Function try to convert tail from direct item into indirect.
** It set up nopack attribute in the inode.u.reiserfs_i.nopack
*/
int reiserfs_unpack (struct inode * inode, struct file * filp)
{
    int retval, windex, exitcode;
    INITIALIZE_PATH(path);
    struct cpu_key key;
    struct buffer_head * bh, * unbh = 0;
    struct item_head * ih;
    loff_t tail_offs;
    unsigned int t1;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3 ;
    struct reiserfs_transaction_handle th ;
    	
    lock_kernel();

    /* check is inode has 'new' statdata */
    if (inode_items_version (inode) == ITEM_VERSION_1)
	return -EINVAL;
	
    journal_begin(&th, inode->i_sb, jbegin_count);
    windex = push_journal_writer("reiserfs_ioctl_unpack");
    exitcode = 0;
    
search_again:
    /* find file's tail */
    t1 = ~(inode->i_sb->s_blocksize - 1);
    make_cpu_key (&key, inode, (inode->i_size & t1) + 1, TYPE_DIRECT, 3);
    retval = search_for_position_by_key (inode->i_sb, &key, &path);

    if (retval == POSITION_NOT_FOUND) {
	if (unbh != NULL) {
	    unsigned long tmp = unbh->b_blocknr;
	    bforget (unbh);
	    /* free what has been allocated by get_new_buffer */
	    reiserfs_free_block (&th, tmp);
	}
    } else {    
	ih = get_ih (&path);
	/* if last item is direct then we have to convert it to indirect */
	if (is_direct_le_ih (ih)) {
	    bh = get_bh (&path);
	    /* allocate new unformatted node to place tail */
	    if (!unbh) {
		retval = get_new_buffer (&th, bh, &unbh, &path);
		if (!unbh) {
		    /* can't allocate block */
		    pathrelse (&path);
		    exitcode = -ENOSPC;
		    goto finish;
		}
		if (retval & SCHEDULE_OCCURRED) {
	    	    /* get_new_buffer was blocked and path had ability to change */
		    pathrelse (&path);
	    	    restart_transaction(&th, inode, &path);
	    	    goto search_again;
		}
	    }
	    tail_offs = ((le_ih_k_offset (ih) - 1) & ~(inode->i_sb->s_blocksize - 1)) + 1;
	    mark_buffer_uptodate (unbh, 1);
	    /* try to convert direct item into indirect */
	    retval = direct2indirect (&th, inode, &path, unbh, tail_offs);
	    if (retval) {		
		/* direct2indirect() did not convert item */
		unsigned long tmp = unbh->b_blocknr;
		bforget (unbh);
		/* free what has been allocated by get_new_buffer */
		reiserfs_free_block (&th, tmp);
		inode->u.reiserfs_i.nopack = 0;
		exitcode = -ENOSPC; /* FIXME: what error has to be returned here? -az */
		goto finish;
	    }
	    brelse (unbh);	    
	}   
	pathrelse (&path);
    }
    /* don't pack tail anymore */
    inode->u.reiserfs_i.nopack = 1;
finish:
    pop_journal_writer(windex);
    journal_end(&th, inode->i_sb, jbegin_count);
    unlock_kernel();    
    return exitcode;
}
